/*
 * Copyright 2016-2017 Golden Code Development Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"),  to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Implementation of ABL procedures which comprise an API exposed to JavaScript. The naming
 * convention for procedures which implement up-calls from JavaScript into converted ABL is to
 * use the prefix "inMsg_".
 *
 * Each such procedure invokes a companion procedure whose name has the prefix "sendMsg_". This
 * is the procedure which performs the actual work of the API. When it is done with its work, the
 * "sendMsg_" procedure invokes a JavaScript callback function based on a unique message ID
 * provided in the originating up-call. Data gathered by the API is passed as a JSON object to
 * the callback function via P2J-REMOTE-CALL.
 *
 * This file contains two examples of APIs which perform database access and "return" row data
 * within JSON objects. It also contains a procedure which is driven by a JavaScript navigation
 * bar to select the embedded ADM2 window to be displayed when the user clicks on a menu item in
 * the navigation bar.
 *
 * TODO: push the up-call/JSON scaffolding code up into the FWD runtime, so that only the core
 * functionality is implemented in the ABL here.
 */

{src/adm2/containr.i}

/* temp-table to count rooms utilized (reserved or occupied) by date */
def temp-table tt-rooms-used
    field use-date as date
    field room-num as int
    field room-type as int
    field rate as dec decimals 2
    index pi-date-room is primary is unique use-date room-num asc
    index si-type room-type asc.
    
/* temp-table of aggregated data for room types over a period */
def temp-table tt-room-type-agg
    field room-type as int
    field description as char
    field count as int
    field potential-revenue as dec decimals 2
    index pi-type is primary is unique room-type asc.
    
DEF VAR hpage AS HANDLE.
def new global shared var wWin as handle.

/*
 * Up-call from JavaScript to notify of client disconnect, delete the current procedure, and quit.
 */
procedure inMsg_notifyEHotelDisconnect.
   delete procedure this-procedure.
   quit.
end.

/*
 * Up-call from JavaScript to gather hotel room utilization data for a range of dates.
 *
 * Parameters:
 *
 *    messageId (input)    - Unique message ID for the up-call.
 *    start-date (input)   - Start of date range.
 *    end-date (input)     - End of date range.
 */
procedure inMsg_utilizationByDate.

   def input param messageId as int.
   def input param start-date as date.
   def input param end-date as date.
   
   run sendMsg_utilizationByDate(messageId, start-date, end-date).
   
end.

/*
 * Collect hotel room utilization data for a range of dates and supply that data back to
 * JavaScript code as a JSON message, via a remote call.
 *
 * Parameters:
 *
 *    messageId (input)    - Unique message ID for the up-call.
 *    start-date (input)   - Start of date range.
 *    end-date (input)     - End of date range.
 */
procedure sendMsg_utilizationByDate.

   def input param messageId as int.
   def input param start-date as date.
   def input param end-date as date.
   
   def var next-date as date no-undo.
   def var d1 as date no-undo.
   def var d2 as date no-undo.
   def var use-count as decimal.
   def var total-rooms as decimal init 0.
   def var full as decimal decimals 4.
   def var jsonMsg as char.
   def var row-data as char.
   def var comma as log init no.
   
   for each room no-lock:
      total-rooms = total-rooms + 1.
   end.
   
   /* collect reservations; include check-in date, exclude check-out date */
   for each reservation
      where reservation.start-date <= end-date
        and reservation.end-date > start-date
      no-lock:
      
      d1 = (if reservation.start-date > start-date then reservation.start-date else start-date).
      d2 = (if reservation.end-date <= end-date then reservation.end-date - 1 else end-date).
      
      do next-date = d1 to d2:
      
         if not can-find(tt-rooms-used
                         where tt-rooms-used.use-date = next-date
                           and tt-rooms-used.room-num = reservation.room-num) then
         do:
            create tt-rooms-used.
            assign
               tt-rooms-used.use-date = next-date
               tt-rooms-used.room-num = reservation.room-num.
         end
         
      end.
      
   end.
   
   /* collect stays; include check-in date, exclude check-out date */
   for each stay
      where stay.start-date <= end-date
        and stay.end-date > start-date
      no-lock:
      
      d1 = (if stay.start-date > start-date then stay.start-date else start-date).
      d2 = (if stay.end-date <= end-date then stay.end-date - 1 else end-date).
      
      do next-date = d1 to d2:
      
         if not can-find(tt-rooms-used
                         where tt-rooms-used.use-date = next-date
                           and tt-rooms-used.room-num = stay.room-num) then
         do:
            create tt-rooms-used.
            assign
               tt-rooms-used.use-date = next-date
               tt-rooms-used.room-num = stay.room-num.
         end
         
      end.
      
   end.
   
   /* count up used rooms by date */
   for each tt-rooms-used use-index pi-date-room
       break by tt-rooms-used.use-date:
      
      if first-of(tt-rooms-used.use-date) then use-count = 0.
      
      use-count = use-count + 1.
      
      if last-of(tt-rooms-used.use-date) then
      do:
      
         /* treat overbookings as fully reserved */
         if (use-count > total-rooms) then use-count = total-rooms.
         
         row-data = row-data + (if comma then ',~n' else '').
         
         full = use-count / total-rooms.
         row-data = row-data + '~{'.
         row-data = row-data + '"date":"' + string(tt-rooms-used.use-date, "99/99/9999") + '",'.
         row-data = row-data + '"score":' + string(full, "z9.9999").
         row-data = row-data + '~}'.
         
         comma = yes.
      end.
      
   end.
   
   empty temp-table tt-rooms-used no-error.
   
   jsonMsg = "~{~n".
   jsonMsg = jsonMsg + ' "api":"utilizationByDateResponse",~n'.
   jsonMsg = jsonMsg + ' "messageId":"' + string(messageId) + '",~n'.
   jsonMsg = jsonMsg + ' "rows": [' + row-data + '],~n'.
   jsonMsg = jsonMsg + ' "domain": [ 0.0, 1.0 ],~n'.
   jsonMsg = jsonMsg + ' "start":"' + string(start-date, "99/99/9999") + '",~n'.
   jsonMsg = jsonMsg + ' "end":"' + string(end-date, "99/99/9999") + '"~n'.
   jsonMsg = jsonMsg + "~}".
   
   P2J-REMOTE-CALL(jsonMsg).
   
end.

/*
 * Up-call from JavaScript to gather hotel room utilization data, grouped by room type, for a
 * range of dates.
 *
 * Parameters:
 *
 *    messageId (input)    - Unique message ID for the up-call.
 *    start-date (input)   - Start of date range.
 *    end-date (input)     - End of date range.
 */
procedure inMsg_utilizationByType.

   def input param messageId as int.
   def input param start-date as date.
   def input param end-date as date.

   run sendMsg_utilizationByType(messageId, start-date, end-date).
end.

/*
 * Collect hotel room utilization data, grouped by room type, for a range of dates. Supply that
 * data back to JavaScript code as a JSON message, via a remote call.
 *
 * Parameters:
 *
 *    messageId (input)    - Unique message ID for the up-call.
 *    start-date (input)   - Start of date range.
 *    end-date (input)     - End of date range.
 */
procedure sendMsg_utilizationByType.

   def input param messageId as int.
   def input param start-date as date.
   def input param end-date as date.
   
   def var jsonMsg as char no-undo.
   def var msg as char init "" no-undo.
   def var weight as dec format "z9.99" no-undo.
   def var score as dec format "z9.99" no-undo.
   def var total-room-count as int init 0 no-undo.
   def var actual-count as dec no-undo.
   def var revenue as dec no-undo.
   def var total-pot-rev as dec decimals 2 init 0 no-undo.
   def var actual-revenue as dec decimals 2 no-undo.
   def var d1 as date no-undo.
   def var d2 as date no-undo.
   def var day-count as int no-undo.
   def var next-date as date no-undo.
   def var comma as log init no no-undo.
   def var rate-weighted as log init yes no-undo.
   def var r1 as dec decimals 2 no-undo.

   /* count the number of active rooms of each type */
   for each room-type no-lock,
       each room no-lock
      where room.room-type = room-type.room-type
        and room.active
      break by room.room-type:
      
       accumulate room.room-num (sub-count by room.room-type).
       if last-of(room.room-type)
       then do:
          create tt-room-type-agg.
          assign tt-room-type-agg.room-type = room.room-type
                 tt-room-type-agg.count = accum sub-count by room.room-type room.room-num
                 tt-room-type-agg.description = room-type.description.
          total-room-count = total-room-count + tt-room-type-agg.count.
       end.
       
   end.
   
   /* get total count of rooms that could potentially be rented for the period */
   day-count = end-date - start-date + 1.
   total-room-count = total-room-count * day-count.
   
   /* calculate potential revenue for the period by room type using the associated room rates */
   /* note: this algorithm assumes rates do not overlap! */
   for each tt-room-type-agg,
       each rate no-lock
      where rate.room-type = tt-room-type-agg.room-type
        and rate.start-date <= end-date
        and rate.end-date >= start-date:
      
      d1 = (if rate.start-date > start-date then rate.start-date else start-date).
      d2 = (if rate.end-date <= end-date then rate.end-date else end-date).
      
      day-count = d2 - d1 + 1.
      revenue = day-count * rate.rate * tt-room-type-agg.count.
      tt-room-type-agg.potential-revenue = tt-room-type-agg.potential-revenue + revenue.
      total-pot-rev = total-pot-rev + revenue.
            
   end.
   
   /* collect daily room utilization data for each reservation for the period */
   for each room no-lock
      where room.active,
       each reservation no-lock
      where reservation.room-num = room.room-num
        and reservation.start-date <= end-date
        and reservation.end-date > start-date:
   
      d1 = (if reservation.start-date > start-date then reservation.start-date else start-date).
      d2 = (if reservation.end-date <= end-date then reservation.end-date - 1 else end-date).
      
      do next-date = d1 to d2:
      
         r1 = 0.
         if rate-weighted
         then do:
            find first rate
                 where rate.room-type = room.room-type
                   and rate.start-date <= next-date
                   and rate.end-date >= next-date
                 no-lock no-error.
            if avail rate
            then r1 = rate.rate.
            else rate-weighted = no.
         end.
         
         create tt-rooms-used.
         assign tt-rooms-used.use-date = next-date
                tt-rooms-used.room-num = reservation.room-num
                tt-rooms-used.room-type = room.room-type
                tt-rooms-used.rate = r1.
         
      end.
      
   end.
   
   /* collect daily room utilization data for each stay for the period */
   for each room no-lock
      where room.active,
       each stay no-lock
      where stay.room-num = room.room-num
        and stay.start-date <= end-date
        and stay.end-date > start-date:
   
      d1 = (if stay.start-date > start-date then stay.start-date else start-date).
      d2 = (if stay.end-date <= end-date then stay.end-date - 1 else end-date).
      
      do next-date = d1 to d2:
      
         r1 = 0.
         if rate-weighted
         then do:
            find first rate
                 where rate.room-type = room.room-type
                   and rate.start-date <= next-date
                   and rate.end-date >= next-date
                 no-lock no-error.
            if avail rate
            then r1 = rate.rate.
            else rate-weighted = no.
         end.
         
         create tt-rooms-used.
         assign tt-rooms-used.use-date = next-date
                tt-rooms-used.room-num = stay.room-num
                tt-rooms-used.room-type = room.room-type
                tt-rooms-used.rate = r1.
         
      end.
      
   end.
   
   /* walk the collected, aggregate data and compose summary data for each room type in JSON */
   for each tt-room-type-agg:
   
      actual-revenue = 0.
      actual-count = 0.
      
      for each tt-rooms-used
         where tt-rooms-used.room-type = tt-room-type-agg.room-type:
      
         actual-revenue = actual-revenue + tt-rooms-used.rate.
         actual-count = actual-count + 1.
         
      end.
      
      if rate-weighted
      then do:
         weight = tt-room-type-agg.potential-revenue / total-pot-rev * 100.
         score = actual-revenue / tt-room-type-agg.potential-revenue * 100.
      end
      else do:
         weight = tt-room-type-agg.count / total-room-count * 100.
         score = (actual-count / (tt-room-type-agg.count * day-count)) * 100.
      end.
      
      msg = msg + (if comma then ",~n" else "").
      msg = msg + "~{".
      msg = msg + '"id":"' + tt-room-type-agg.room-type + '",'.
      msg = msg + '"score":"' + string(score, "zz9.99") + '",'.
      msg = msg + '"weight":"' + string(weight, "zz9.99") + '",'.
      msg = msg + '"actualRevenue":"' + actual-revenue + '",'.
      msg = msg + '"potentialRevenue":"' + tt-room-type-agg.potential-revenue + '",'.
      msg = msg + '"roomsOccupied":"' + string(actual-count, "zzzzz9") + '",'.
      msg = msg + '"totalRooms":"' + string(tt-room-type-agg.count * day-count, "zzzzz9") + '",'.
      msg = msg + '"label":"' + tt-room-type-agg.description + '"'.
      msg = msg + "~}".
      
      comma = yes.
      
   end.
       
   jsonMsg = "~{~n".
   jsonMsg = jsonMsg + ' "api":"utilizationByTypeResponse",~n'.
   jsonMsg = jsonMsg + ' "messageId":"' + string(messageId) + '",~n'.
   jsonMsg = jsonMsg + ' "rows": [~n' + msg + '],~n'.
   jsonMsg = jsonMsg + ' "rateWeighted":"' + (if rate-weighted then "true" else "false") + '",~n'.
   jsonMsg = jsonMsg + ' "start":"' + string(start-date, "99/99/9999") + '",~n'.
   jsonMsg = jsonMsg + ' "end":"' + string(end-date, "99/99/9999") + '"~n'.
   jsonMsg = jsonMsg + "~}".
   
   empty temp-table tt-room-type-agg.
   empty temp-table tt-rooms-used.
   
   P2J-REMOTE-CALL(jsonMsg).
   
end.

/*
 * Display an embedded ADM2 window in an iframe.
 *
 * Parameters:
 *
 *    pname (input)  -  Name of external procedure which loads the desired window.
 */
procedure showEmbeddedWindow.
    def input param pname as char.
    
    RUN constructObject (
          INPUT  pname ,
          INPUT  wWin,
          INPUT  'HideOnInitnoDisableOnInitnoObjectLayout':U ,
          OUTPUT hpage ).
    RUN repositionObject IN hpage (1, 1) NO-ERROR.
    RUN initializeObject IN hpage.
end.

/*
 * Up-call from JavaScript which executes a command to display a particular, embedded ADM2 window
 * in an iframe.
 *
 * Parameters:
 *
 *    messageId (input)    -  Unique message ID for the up-call.
 *    command (input)      -  Identifying string associated with the external procedure which
 *                            loads the window.
 */
procedure inMsg_executeCommand.
   def input param messageId as int.
   def input param command as char.

   if hpage <> ? then do:
      run destroyObject in hpage.
   end.

   case command:
      when "showAvailRooms" then do:
         run showEmbeddedWindow("avail-rooms-frame.w").
      end.
      when "showGuests" then do:
         run showEmbeddedWindow("guests-frame.w").
      end.
      when "showReservations" then do:
         run showEmbeddedWindow("reservations-frame.w").
      end.
      when "showRates" then do:
         run showEmbeddedWindow("rates-frame.w").
      end.
      when "showRooms" then do:
         run showEmbeddedWindow("rooms-frame.w").
      end.
      otherwise do:
         message "Invalid command:" command view-as alert-box.
      end.
   end.
end.
